from ...helpers import signature
from .l01x0_common_layers import Layer
from textual.message_pump import MessagePump
from textual.widget import Widget
import asyncio
import types
import copy
import inspect
import functools

class textualinit(signature):
    default = ('', {})

'''
class TextualShellLayer(Layer):
    class GeneratorOverride:
        def __init__(self, scope, target):
            functools.update_wrapper(self,target)
            self.scope=scope
            self.target=target
            sig = inspect.signature(self) #sig = inspect.signature(target)
            params = list(sig.parameters.items())
            params.pop(1)
            sig1 = sig.replace(parameters=[b for a,b in params])
            self.__signature__ = sig1
            for k,v in target.__dict__.items():
                if k.startswith('_textual'):
                    setattr(self,k,v)
            #self.__positional_argument_count__ = sum(1 for param in sig1.parameters.values() if param._default == inspect._empty)-1
            self.__positional_argument_count__ = self.target.__code__.co_argcount - 2
            if self.target.__defaults__:
                self.__positional_argument_count__ -= len(self.target.__defaults__)
            pass
        def __get__(self,obj,cls):
            return types.MethodType(self,obj)     
        def __call__(self,instance,*args,**kwargs):
            yield from self.target(instance,self.scope,*args[:self.__positional_argument_count__],**kwargs)
    class MethodOverride:
        def __init__(self, scope, target):
            functools.update_wrapper(self,target)
            self.scope=scope
            self.target=target
            sig = inspect.signature(self) #sig = inspect.signature(target)
            params = list(sig.parameters.items())
            params.pop(1)
            sig1 = sig.replace(parameters=[b for a,b in params])
            self.__signature__ = sig1
            for k,v in target.__dict__.items():
                if k.startswith('_textual'):
                    setattr(self,k,v)
            #self.__positional_argument_count__ = sum(1 for param in sig1.parameters.values() if param._default == inspect._empty)-1
            self.__positional_argument_count__ = self.target.__code__.co_argcount - 2
            if self.target.__defaults__:
                self.__positional_argument_count__ -= len(self.target.__defaults__)
            pass
        def __get__(self,obj,cls):
            return types.MethodType(self,obj)     
        def __call__(self,instance,*args,**kwargs):
            return self.target(instance,self.scope,*args[:self.__positional_argument_count__],**kwargs)


    def item_build(self, builder):
        scope = builder.scope
        this = None 
        if builder.type is not None and (issubclass(builder.type, Widget) or type(builder.type) is type):
            init = builder.kwargs.get('init', None)
            if init is not None:
                if type(init) is textualinit:
                    method, argp, args = init 
                elif type(init) is tuple:
                    method, args = init 
                elif type(init) is dict:
                    args = init
                this = builder.type(**args)
            else:
                this = builder.type()
        else:
            if isinstance(builder.item, types.FunctionType):
                pass
            elif issubclass(builder.item, Widget):
                def Caller_fn(self):
                    pass
                class Caller:
                    def __init__(self):
                        functools.update_wrapper(self,Caller_fn)
                        #self.__signature__ = inspect.signature(Caller_fn).replace()
                        #self.__qualname__ = 'Caller'
                        pass
                    def __get__(self,obj,cls):
                        return types.MethodType(self,obj)     
                    def __call__(self, *args, **kwds):
                        pass
                class WidgetOverride(builder.item):
                    def __init__(self, *args, **kwargs):
                        super().__init__(*args, **kwargs)
                    aaa_caller = Caller()
                    pass
                #WidgetOverride = builder.item

                all_methods = inspect.getmembers(builder.item, predicate=inspect.isfunction)                    
                # Get methods defined in the class itself
                class_methods = {name: method for name, method in all_methods if method.__qualname__.startswith(builder.item.__qualname__ + '.')}

                dh = builder.item.__dict__.get('_decorated_handlers',None)
                if dh:
                    dh=copy.deepcopy(dh)
                for method_name in class_methods:
                    method = getattr(WidgetOverride,method_name,None)
                    if method is not None:
                        if method_name.startswith('_'):
                            continue
                        if method_name in ('compose'):
                            wrapper = TextualShellLayer.GeneratorOverride(scope,method)
                        else:
                            wrapper = TextualShellLayer.MethodOverride(scope,method)
                        setattr(WidgetOverride,method_name,wrapper)                       
                        if dh:
                            for k,v in dh.items():
                                for i, h in enumerate(v):
                                    m, _ = h
                                    if m is method:
                                        v[i] = (wrapper,_)
                if dh:
                    setattr(WidgetOverride,'_decorated_handlers',dh)
                
                init = builder.kwargs.get('init', None)
                if init is not None:
                    if type(init) is tuple:
                        method, args = init 
                    elif type(init) is dict:
                        args = init
                    this = WidgetOverride(**args)
                else:
                    this = WidgetOverride()
                pass
            elif type(builder.item) is type:
                init = builder.kwargs.get('init', None)
                if init is not None:
                    if type(init) is tuple:
                        method, args = init 
                    elif type(init) is dict:
                        args = init
                    this = builder.item(**args)
                else:
                    this = builder.item()
        if this is None:
            this = builder.item
        if this is not None:
            if isinstance(this,Widget):
                scope.__joint__.tool = this
            else:
                scope.__joint__.mass = this  
        return this
'''

class TextualShellLayer(Layer):
    def item_build(self, builder):
        scope = builder.scope
        this = None 
        if builder.type is not None and (issubclass(builder.type, MessagePump) or type(builder.type) is type):
            init = builder.kwargs.get('init', None)
            if init is not None:
                if type(init) is textualinit:
                    method, argp, args = init 
                elif type(init) is tuple:
                    method, args = init 
                elif type(init) is dict:
                    args = init
                this = builder.type(**args)
            else:
                this = builder.type()
        else:
            if isinstance(builder.item, types.FunctionType):
                pass
            elif issubclass(builder.item, MessagePump):
                init = builder.kwargs.get('init', None)
                if init is not None:
                    if type(init) is tuple:
                        method, args = init 
                    elif type(init) is dict:
                        args = init
                    this = builder.item(**args)
                else:
                    this = builder.item()
                pass
            elif type(builder.item) is type:
                init = builder.kwargs.get('init', None)
                if init is not None:
                    if type(init) is tuple:
                        method, args = init 
                    elif type(init) is dict:
                        args = init
                    this = builder.item(**args)
                else:
                    this = builder.item()
        if this is None:
            this = builder.item
        if this is not None:
            if isinstance(this,MessagePump):
                scope.__joint__.tool = this
            else:
                scope.__joint__.mass = this  
        return this

class TextualCompositionLayer(TextualShellLayer):
    def __init__(self):
        self.build_list = []
        self.stack = []
        self.composition = True
        self.last_list_id = 0
    def __enter__(self):
        self.composition = True
        return super().__enter__()
    def item_build(self, builder):
        e = super().item_build(builder)
        if e is builder.scope.__joint__.tool:
           self.build_list.append(e)
        return e
    def item_enter(self, builder):
        if builder.this is builder.scope.__joint__.tool:
            self.stack.append(self.build_list)        
            self.build_list = []
    def item_leave(self, builder):
        if builder.this is builder.scope.__joint__.tool:
            l = self.build_list
            self.build_list = self.stack.pop()
            if l:
                self.build_list[-1] = (self.build_list[-1],l)
    def drop(self):
        blist = self.build_list
        self.build_list = []
        self.composition = False
        def walker(l):
            for x in l:
                if type(x) is tuple:
                    xi, xl = x
                    with xi:
                        yield from walker(xl)
                else:
                    yield x
        yield from walker(blist)

    def __iter__(self):
        yield from self.drop()              

    def item_stage(self, builder):
        if self.composition:
            return
        if id(self.build_list) == self.last_list_id:
            return
        pass
        self.last_list_id = id(self.build_list)
        widget = builder.scope.__joint__.this
        container = builder.scope.__joint__.base.__joint__.this
        loop : asyncio.AbstractEventLoop  = container.screen.app._loop
        async def mount(self, builder, container:Widget):
            async def walker(c:Widget, l):
                for x in l:
                    if type(x) is tuple:
                        xi, xl = x
                        result1 = await c.mount(xi)                        
                        result2 = await walker(xi, xl)
                    else:
                        result3 = await c.mount(x)
            blist = self.build_list
            self.build_list = []
            result = await walker(container, blist)
            
        #loop.call_soon_threadsafe(mount, self, builder, container, widget)
        future = asyncio.run_coroutine_threadsafe(mount(self, builder, container),loop)
        ## result = future.result()  # questo fucnziona solo se il thread del loop non è il thread corrente, cosa che invece pare che sia
        pass


  



from pyrite.attract import locate
import clipboard as cb
from sys import argv
import argparse

def main():
    parser = argparse.ArgumentParser("attract-locate.py")
    parser.add_argument("item", help="Item name fragment", type=str)
    g0 = parser.add_mutually_exclusive_group()
    g0.add_argument("--verbose", "-v", help="Verbose output", action='store_true')
    g0.add_argument("--silent", "-s", help="Silent. No output to stdout", action='store_true')
    parser.add_argument("--clipboard", "-c", help="Copy results to clipboard", action='store_true')
    parser.add_argument("--resolve-links", "-r", help="Resolve links", action='store_true')
    g1 = parser.add_mutually_exclusive_group()
    g1.add_argument("--posix-path", "-p", help="Posix pathname", action='store_true')
    g1.add_argument("--unc-path", "-u", help="UNC pathname", action='store_true')
    g1.add_argument("--wsl-path", "-w", help="WSL pathname", action='store_true')
    args = parser.parse_args()
    result = locate(args.item,options=['resolve_links' if args.resolve_links else None]) 
    if result is None:
        if not args.silent:
            if args.verbose:
                print(f'[E1]attract-locate: "{args.item}" not found.')
        exit(1)
    if args.posix_path:
        result = result.replace('\\','/')    
    elif args.unc_path:
        result = result.replace('/','\\')    
    elif args.wsl_path:
        result = result.replace("\\","/")
        result = f'/mnt/{result[0].lower()}{result[2:]}'
    if not args.silent:
        if args.verbose:
            print(f'[OK]attract-locate: "{result}" found with "{args.item}"')
        print(result)
    if args.clipboard:
        cb.copy(result)

#main()


from .gui12x0_wintypes import *
from . import gui13x0_winapi as winapi

class PropertyStore:
    stock_properties_catalog = PropertyStock.catalog
    def __init__(self, target=None):
        self.target = ctypes.POINTER(IPropertyStore)()
        self.invoke = None
    def __del__(self):
        if self.target and self.invoke:
            self.invoke.Release(self.target)  
    def set_value(self, prop, value, options=None):
        prop_key = PROPERTYKEY()
        prop_variant = PROPVARIANT()
        prop_key.fmtid, prop_key.pid, py_type, prop_variant.vt = self.stock_properties_catalog[prop]
        if value is not None:
            match(prop_variant.vt):
                case PROPVARIANT.VT_LPWSTR:
                    prop_variant.vVal.pwszVal = str(value)
                case PROPVARIANT.VT_BOOL:
                    prop_variant.vVal.boolVal = wintraits.TRUE if bool(value) else wintraits.FALSE 
                case _:
                    raise Exception('ERROR')
        else:
            prop_variant.vt = PROPVARIANT.VT_EMPTY
        result = self.invoke.SetValue(self.target,prop_key, prop_variant)
        if result != 0:
            raise Exception('ERROR')
    def get_value(self,prop, options=None):
        prop_key = PROPERTYKEY()
        prop_variant = PROPVARIANT()
        prop_key.fmtid, prop_key.pid, py_type, var_type = self.stock_properties_catalog[prop]
        result = self.invoke.GetValue(self.target,prop_key, prop_variant)
        if result != 0:
            raise Exception('ERROR')
        match(prop_variant.vt):
            case PROPVARIANT.VT_LPWSTR:
                value = str(prop_variant.vVal.pwszVal)
            case PROPVARIANT.VT_BOOL:
                value = True if prop_variant.vVal.boolVal != 0 else False
            case PROPVARIANT.VT_EMPTY:
                value = None
            case _:
                raise Exception('ERROR')
        return value
    def commit(self):
        result = self.invoke.Commit(self.target)
        return True if result == 0 else False
    def bind(self, target=None):
        if self.target != target:
            if self.target:
                if self.invoke:
                    self.invoke.Release(self.target)
                    self.invoke = None
            self.target = target
        if self.target:
            if not self.invoke:
                self.invoke = target.contents.lpVtbl.contents
        return self
    def __len__(self):
        length = ctypes.c_ulong()
        result = self.invoke.GetCount(self.target,ctypes.byref(length))
        if result != 0:
            raise Exception('ERROR')
        return int(length)
    def __getitem__(self, key):
        return self.get_value(key)
    def __setitem__(self, key, value):
        return self.set_value(key,value)

    def com_iid(self):
        return IPropertyStore.IID_IPropertyStore
    def com_ptr(self):
        return self.target
    
    @staticmethod
    def from_hwnd(hwnd):
        ps = PropertyStore()
        result = winapi.SHGetPropertyStoreForWindow(int(hwnd), ps.com_iid(), ps.com_ptr())
        if result != 0:
            return None
        return ps.bind(ps.com_ptr())




#import win32api, win32con, win32gui, win32process, win32ui
from . import gui11x0_wintraits as wintraits
from . import gui12x0_wintypes as wintypes

from win32api import GetMonitorInfo, MonitorFromWindow, MonitorFromRect, \
                        EnumDisplayMonitors, GetAsyncKeyState, \
                        WinExec, CloseHandle, OpenProcess, RGB, \
                        GetCurrentThreadId
from win32gui import GetWindowLong, SetWindowLong, GetForegroundWindow, \
                        GetWindowText, GetClassName, GetClassLong, IsIconic, IsWindowVisible, IsWindow, \
                        EnumWindows, EnumChildWindows, GetWindowRect, GetParent, ShowWindow, MoveWindow, \
                        GetWindow, SetFocus, GetFocus, \
                        PostMessage, PostQuitMessage, SendMessage, GetIconInfo, GetObject, SetWindowPos, \
                        RegisterHotKey, SetWindowLong, GetWindowLong, GetActiveWindow,\
                        SetParent, SetLayeredWindowAttributes, \
                        GetForegroundWindow, SetForegroundWindow,\
                        GetActiveWindow, SetActiveWindow, \
                        SetCapture, GetCapture
from win32ui import CreateBitmapFromHandle
from win32process import GetWindowThreadProcessId, GetCurrentProcessId, AttachThreadInput

import ctypes

user32 = ctypes.windll.user32
kernel32 = ctypes.windll.kernel32
dwmapi = ctypes.windll.dwmapi
shell32 = ctypes.windll.shell32
ole32 = ctypes.windll.ole32

__CoInitialize = ole32.CoInitialize
__CoInitialize.restype = ctypes.HRESULT
__CoInitialize.argtypes = [ctypes.c_void_p]
def CoInitialize(arg=None):
    __CoInitialize(wintraits.NULL)
CoUninitialize = ole32.CoUninitialize
CoUninitialize.restype = None
CoUninitialize.argtypes = None

SHGetPropertyStoreForWindow = shell32.SHGetPropertyStoreForWindow
SHGetPropertyStoreForWindow.restype = ctypes.HRESULT
SHGetPropertyStoreForWindow.argtypes = [wintypes.HWND, ctypes.POINTER(wintypes.GUID), ctypes.POINTER(ctypes.POINTER(wintypes.IPropertyStore))]

__GetDpiForWindow = user32.GetDpiForWindow
__GetClassLongPtr = user32.GetClassLongPtrA
__SetProcessDpiAwarenessContext = user32.SetProcessDpiAwarenessContext


def GetDpiForWindow(hwnd):
    return __GetDpiForWindow(hwnd)

def GetClassLongPtr(hwnd, index):
    return __GetClassLongPtr(hwnd, ctypes.c_int(index))

def SetProcessDpiAwarenessContext(index):
    return __SetProcessDpiAwarenessContext(wintypes.HANDLE(index))


# Define constants
__DWM_EC_DISABLECOMPOSITION = 0
__DWM_EC_ENABLECOMPOSITION = 1
# Define the attribute you want to get
__DWMWA_EXTENDED_FRAME_BOUNDS = 9

# Define the DwmGetWindowAttribute function
__DwmGetWindowAttribute = dwmapi.DwmGetWindowAttribute
__DwmGetWindowAttribute.argtypes = [wintypes.HWND, ctypes.c_uint, ctypes.POINTER(ctypes.c_void_p), ctypes.c_uint]
__DwmGetWindowAttribute.restype = ctypes.HRESULT

def DwmGetWindowExtendedFrameBounds(hwnd):
    rect = wintypes.RECT()
    rect_ptr = ctypes.pointer(rect)
    void_ptr = ctypes.cast( rect_ptr, ctypes.POINTER(ctypes.c_void_p) )
    result = __DwmGetWindowAttribute(hwnd, __DWMWA_EXTENDED_FRAME_BOUNDS, void_ptr, ctypes.sizeof(rect))
    if result != 0:
        raise ctypes.WinError(result)
    return rect.left, rect.top, rect.right, rect.bottom

CreateProcess = kernel32.CreateProcessW
CreateProcess.argtypes = [
    wintypes.LPCWSTR,                               # lpApplicationName,
    wintypes.LPWSTR,                                # lpCommandLine,
    ctypes.POINTER(ctypes.c_void_p),                # lpProcessAttributes
    ctypes.POINTER(ctypes.c_void_p),                # lpThreadAttributes
    wintypes.BOOL,                                  # bInheritHandles
    wintypes.DWORD,                                 # dwCreationFlags
    ctypes.POINTER(ctypes.c_void_p),                # lpEnvironment
    wintypes.LPCWSTR,                               # lpCurrentDirectory
    ctypes.POINTER(wintypes.STARTUPINFOEX),         # lpStartupInfo
    ctypes.POINTER(wintypes.PROCESS_INFORMATION)    # lpProcessInformation
]
CreateProcess.restype = wintypes.BOOL

GetShellWindow = user32.GetShellWindow
GetShellWindow.argtypes = []
GetShellWindow.restype = wintypes.HWND

# Define the function prototype
InitializeProcThreadAttributeList = kernel32.InitializeProcThreadAttributeList
InitializeProcThreadAttributeList.argtypes = [
    ctypes.POINTER(ctypes.c_void_p),  # lpAttributeList
    wintypes.DWORD,                   # dwAttributeCount
    wintypes.DWORD,                   # dwFlags
    ctypes.POINTER(ctypes.c_size_t)   # lpSize
]
InitializeProcThreadAttributeList.restype = wintypes.BOOL

UpdateProcThreadAttribute = kernel32.UpdateProcThreadAttribute
UpdateProcThreadAttribute.argtypes = [
    ctypes.POINTER(ctypes.c_void_p),    # lpAttributeList
    wintypes.DWORD,                     # dwFlags
    ctypes.POINTER(wintypes.DWORD),     # Attribute
    ctypes.POINTER(ctypes.c_void_p),    # lpValue
    ctypes.c_size_t,                    # cbSize
    ctypes.POINTER(ctypes.c_void_p),    # lpPreviousValue
    ctypes.POINTER(ctypes.c_size_t)     # lpReturnSize
]
UpdateProcThreadAttribute.restype = wintypes.BOOL

GetLastError = kernel32.GetLastError
GetLastError.argtypes = []
GetLastError.restype = wintypes.DWORD


SetCurrentProcessExplicitAppUserModelID = shell32.SetCurrentProcessExplicitAppUserModelID
SetCurrentProcessExplicitAppUserModelID.argtypes = [wintypes.LPCWSTR]
SetCurrentProcessExplicitAppUserModelID.restype = wintypes.HANDLE


GetConsoleWindow = kernel32.GetConsoleWindow
GetConsoleWindow.argtypes = []
GetConsoleWindow.restype = wintypes.HWND


# Define SetWindowLongPtrW
# example usage of GWL_WNDPROC: SetWindowLongPtr(hwnd, GWL_WNDPROC, ctypes.cast(wnd_proc_callback, ctypes.c_void_p))
SetWindowLongPtr = user32.SetWindowLongPtrW
SetWindowLongPtr.argtypes = [wintypes.HWND, wintypes.INT, ctypes.c_void_p]
SetWindowLongPtr.restype = ctypes.c_void_p

# Define GetWindowLongPtrW
GetWindowLongPtr = user32.GetWindowLongPtrW
GetWindowLongPtr.argtypes = [wintypes.HWND, wintypes.INT]
GetWindowLongPtr.restype = ctypes.c_void_p

# Define CallWindowProcW
CallWindowProc = user32.CallWindowProcW
CallWindowProc.argtypes = [
    ctypes.c_void_p,
    wintypes.HWND,
    wintypes.UINT,
    wintypes.WPARAM,
    wintypes.LPARAM,
]
CallWindowProc.restype = ctypes.c_long


SetConsoleMode = kernel32.SetConsoleMode
SetConsoleMode.argtypes = [wintypes.HANDLE, wintypes.DWORD]
SetConsoleMode.restype = wintypes.BOOL
GetConsoleMode = kernel32.GetConsoleMode
GetConsoleMode.argtypes = [wintypes.HANDLE, wintypes.LPDWORD]
GetConsoleMode.restype = wintypes.BOOL

GetStdHandle = kernel32.GetStdHandle
GetStdHandle.argtypes = [wintypes.DWORD]
GetStdHandle.restype = wintypes.HANDLE
SetStdHandle = kernel32.SetStdHandle
SetStdHandle.argtypes = [wintypes.DWORD, wintypes.HANDLE]
SetStdHandle.restype = wintypes.BOOL
